@@ -4,22 +4,25 @@ Collection of helper functions
 """
 import copy
 import glob
+import logging
 import os
 import os.path
-import errno
-import shutil
 import random
-import logging
+import shutil
 from typing import Optional, List
-import numpy as np
+from pathlib import Path
 import pkg_resources
+import yaml
+
+import numpy as np
 
 import torch
 from torch import nn, Tensor
 from torch.utils.tensorboard import SummaryWriter
 
-from torchtext.data import Dataset
-import yaml
+# pylint: disable=no-name-in-module
+from torchtext.legacy.data import Dataset
+
 from joeynmt.vocabulary import Vocabulary
 from joeynmt.plotting import plot_heatmap
 
@@ -54,7 +57,7 @@ def make_logger(log_dir: str = None, mode: str = "train") -> str:
     :param mode: log file name. 'train', 'test' or 'translate'
     :return: joeynmt version number
     """
-    logger = logging.getLogger("") # root logger
+    logger = logging.getLogger("")  # root logger
     version = pkg_resources.require("joeynmt")[0].version
 
     # add handlers only once.
@@ -96,7 +99,7 @@ def log_cfg(cfg: dict, prefix: str = "cfg") -> None:
             log_cfg(v, prefix=p)
         else:
             p = '.'.join([prefix, k])
-            logger.info("{:34s} : {}".format(p, v))
+            logger.info("%34s : %s", p, v)
 
 
 def clones(module: nn.Module, n: int) -> nn.ModuleList:
@@ -118,8 +121,8 @@ def subsequent_mask(size: int) -> Tensor:
     :param size: size of mask (2nd and 3rd dim)
     :return: Tensor with 0s and 1s of shape (1, size, size)
     """
-    mask = np.triu(np.ones((1, size, size)), k=1).astype('uint8')
-    return torch.from_numpy(mask) == 0
+    ones = torch.ones(size, size, dtype=torch.bool)
+    return torch.tril(ones, out=ones).unsqueeze(0)
 
 
 def set_seed(seed: int) -> None:
@@ -148,19 +151,20 @@ def log_data_info(train_data: Dataset, valid_data: Dataset, test_data: Dataset,
     :param trg_vocab:
     """
     logger = logging.getLogger(__name__)
-    logger.info(
-        "Data set sizes: \n\ttrain %d,\n\tvalid %d,\n\ttest %d",
-            len(train_data), len(valid_data),
-            len(test_data) if test_data is not None else 0)
+    logger.info("Data set sizes: \n\ttrain %d,\n\tvalid %d,\n\ttest %d",
+                len(train_data), len(valid_data),
+                len(test_data) if test_data is not None else 0)
 
     logger.info("First training example:\n\t[SRC] %s\n\t[TRG] %s",
-        " ".join(vars(train_data[0])['src']),
-        " ".join(vars(train_data[0])['trg']))
+                " ".join(vars(train_data[0])['src']),
+                " ".join(vars(train_data[0])['trg']))
 
-    logger.info("First 10 words (src): %s", " ".join(
-        '(%d) %s' % (i, t) for i, t in enumerate(src_vocab.itos[:10])))
-    logger.info("First 10 words (trg): %s", " ".join(
-        '(%d) %s' % (i, t) for i, t in enumerate(trg_vocab.itos[:10])))
+    logger.info(
+        "First 10 words (src): %s",
+        " ".join(f"({i}) {t}" for i, t in enumerate(src_vocab.itos[:10])))
+    logger.info(
+        "First 10 words (trg): %s",
+        " ".join(f"({i}) {t}" for i, t in enumerate(trg_vocab.itos[:10])))
 
     logger.info("Number of Src words (types): %d", len(src_vocab))
     logger.info("Number of Trg words (types): %d", len(trg_vocab))
@@ -173,7 +177,7 @@ def load_config(path="configs/default.yaml") -> dict:
     :param path: path to YAML configuration file
     :return: configuration dictionary
     """
-    with open(path, 'r') as ymlfile:
+    with open(path, 'r', encoding="utf-8") as ymlfile:
         cfg = yaml.safe_load(ymlfile)
     return cfg
 
@@ -189,15 +193,21 @@ def bpe_postprocess(string, bpe_type="subword-nmt") -> str:
     if bpe_type == "sentencepiece":
         ret = string.replace(" ", "").replace("▁", " ").strip()
     elif bpe_type == "subword-nmt":
+        # Remove merge markers within the sentence.
         ret = string.replace("@@ ", "").strip()
+        # Remove final merge marker.
+        if ret.endswith("@@"):
+            ret = ret[:-2]
     else:
         ret = string.strip()
     return ret
 
 
-def store_attention_plots(attentions: np.array, targets: List[List[str]],
+def store_attention_plots(attentions: np.array,
+                          targets: List[List[str]],
                           sources: List[List[str]],
-                          output_prefix: str, indices: List[int],
+                          output_prefix: str,
+                          indices: List[int],
                           tb_writer: Optional[SummaryWriter] = None,
                           steps: int = 0) -> None:
     """
@@ -215,25 +225,31 @@ def store_attention_plots(attentions: np.array, targets: List[List[str]],
     for i in indices:
         if i >= len(sources):
             continue
-        plot_file = "{}.{}.pdf".format(output_prefix, i)
+        plot_file = f"{output_prefix}.{i}.pdf"
         src = sources[i]
         trg = targets[i]
         attention_scores = attentions[i].T
         try:
-            fig = plot_heatmap(scores=attention_scores, column_labels=trg,
-                               row_labels=src, output_path=plot_file,
+            fig = plot_heatmap(scores=attention_scores,
+                               column_labels=trg,
+                               row_labels=src,
+                               output_path=plot_file,
                                dpi=100)
             if tb_writer is not None:
                 # lower resolution for tensorboard
-                fig = plot_heatmap(scores=attention_scores, column_labels=trg,
-                                   row_labels=src, output_path=None, dpi=50)
-                tb_writer.add_figure("attention/{}.".format(i), fig,
+                fig = plot_heatmap(scores=attention_scores,
+                                   column_labels=trg,
+                                   row_labels=src,
+                                   output_path=None,
+                                   dpi=50)
+                tb_writer.add_figure(f"attention/{i}.",
+                                     fig,
                                      global_step=steps)
         # pylint: disable=bare-except
         except:
-            print("Couldn't plot example {}: src len {}, trg len {}, "
-                  "attention scores shape {}".format(i, len(src), len(trg),
-                                                     attention_scores.shape))
+            print(f"Couldn't plot example {i}: src len {len(src)}, "\
+                    f"trg len {len(trg)}, attention scores shape "\
+                    f"{attention_scores.shape}")
             continue
 
 
@@ -245,15 +261,15 @@ def get_latest_checkpoint(ckpt_dir: str) -> Optional[str]:
     :param ckpt_dir:
     :return: latest checkpoint file
     """
-    list_of_files = glob.glob("{}/*.ckpt".format(ckpt_dir))
+    list_of_files = glob.glob(f"{ckpt_dir}/*.ckpt")
     latest_checkpoint = None
     if list_of_files:
         latest_checkpoint = max(list_of_files, key=os.path.getctime)
 
     # check existence
     if latest_checkpoint is None:
-        raise FileNotFoundError("No checkpoint found in directory {}."
-                                .format(ckpt_dir))
+        raise FileNotFoundError(
+            f"No checkpoint found in directory {ckpt_dir}.")
     return latest_checkpoint
 
 
@@ -265,7 +281,7 @@ def load_checkpoint(path: str, use_cuda: bool = True) -> dict:
     :param use_cuda: using cuda or not
     :return: checkpoint (dict)
     """
-    assert os.path.isfile(path), "Checkpoint %s not found" % path
+    assert os.path.isfile(path), f"Checkpoint {path} not found"
     checkpoint = torch.load(path, map_location='cuda' if use_cuda else 'cpu')
     return checkpoint
 
@@ -313,12 +329,68 @@ def freeze_params(module: nn.Module) -> None:
         p.requires_grad = False
 
 
-def symlink_update(target, link_name):
+def delete_ckpt(to_delete: Path) -> None:
+    """
+    Delete checkpoint
+
+    :param to_delete: checkpoint file to be deleted
+    """
+    logger = logging.getLogger(__name__)
     try:
-        os.symlink(target, link_name)
-    except FileExistsError as e:
-        if e.errno == errno.EEXIST:
-            os.remove(link_name)
-            os.symlink(target, link_name)
-        else:
-            raise e
+        logger.info('delete %s', to_delete.as_posix())
+        to_delete.unlink()
+
+    except FileNotFoundError as e:
+        logger.warning(
+            "Wanted to delete old checkpoint %s but "
+            "file does not exist. (%s)", to_delete, e)
+
+
+def symlink_update(target: Path, link_name: Path) -> Optional[Path]:
+    """
+    This function finds the file that the symlink currently points to, sets it
+    to the new target, and returns the previous target if it exists.
+
+    :param target: A path to a file that we want the symlink to point to.
+                    no parent dir, filename only, i.e. "10000.ckpt"
+    :param link_name: This is the name of the symlink that we want to update.
+                    link name with parent dir, i.e. "models/my_model/best.ckpt"
+
+    :return:
+        - current_last: This is the previous target of the symlink, before it is
+            updated in this function. If the symlink did not exist before or did
+            not have a target, None is returned instead.
+    """
+    if link_name.is_symlink():
+        current_last = link_name.resolve()
+        link_name.unlink()
+        link_name.symlink_to(target)
+        return current_last
+    link_name.symlink_to(target)
+    return None
+
+
+def expand_reverse_index(reverse_index: List[int], n_best: int = 1) \
+        -> List[int]:
+    """
+    expand resort_reverse_index for n_best prediction
+
+    ex. 1) reverse_index = [1, 0, 2] and n_best = 2, then this will return
+    [2, 3, 0, 1, 4, 5].
+
+    ex. 2) reverse_index = [1, 0, 2] and n_best = 3, then this will return
+    [3, 4, 5, 0, 1, 2, 6, 7, 8]
+
+    :param reverse_index: reverse_index returned from batch.sort_by_src_length()
+    :param n_best:
+    :return: expanded sort_reverse_index
+    """
+    if n_best == 1:
+        return reverse_index
+
+    resort_reverse_index = []
+    for ix in reverse_index:
+        for n in range(0, n_best):
+            resort_reverse_index.append(ix * n_best + n)
+    assert len(resort_reverse_index) == len(reverse_index) * n_best
+    return resort_reverse_index
