@@ -13,7 +13,7 @@ __all__ = ["greedy", "transformer_greedy", "beam_search", "run_batch"]
 
 
 def greedy(src_mask: Tensor, max_output_length: int, model: Model,
-           encoder_output: Tensor, encoder_hidden: Tensor)\
+           encoder_output: Tensor, encoder_hidden: Tensor) \
         -> (np.array, np.array):
     """
     Greedy decoding. Select the token word highest probability at each time
@@ -137,7 +137,7 @@ def transformer_greedy(
         with torch.no_grad():
             logits, _, _, _ = model(
                 return_type="decode",
-                trg_input=ys, # model.trg_embed(ys) # embed the previous tokens
+                trg_input=ys,  # model.trg_embed(ys) # embed the previous tokens
                 encoder_output=encoder_output,
                 encoder_hidden=None,
                 src_mask=src_mask,
@@ -170,9 +170,7 @@ def beam_search(model: Model, size: int,
     """
     Beam search with size k.
     Inspired by OpenNMT-py, adapted for Transformer.
-
     In each decoding step, find the k most likely partial hypotheses.
-
     :param model:
     :param size: size of the beam
     :param encoder_output:
@@ -185,8 +183,8 @@ def beam_search(model: Model, size: int,
         - stacked_output: output hypotheses (2d array of indices),
         - stacked_attention_scores: attention scores (3d array)
     """
-    assert size > 0, 'Beam size must be >0.'
-    assert n_best <= size, 'Can only return {} best hypotheses.'.format(size)
+    assert size > 0, "Beam size must be >0."
+    assert n_best <= size, f"Can only return {size} best hypotheses."
 
     # init
     bos_index = model.bos_index
@@ -197,8 +195,8 @@ def beam_search(model: Model, size: int,
     transformer = isinstance(model.decoder, TransformerDecoder)
     batch_size = src_mask.size(0)
     att_vectors = None  # not used for Transformer
-    hidden = None       # not used for Transformer
-    trg_mask = None     # not used for RNN
+    hidden = None  # not used for Transformer
+    trg_mask = None  # not used for RNN
 
     # Recurrent models only: initialize RNN hidden state
     # pylint: disable=protected-access
@@ -258,7 +256,12 @@ def beam_search(model: Model, size: int,
         "scores": [[] for _ in range(batch_size)],
         "gold_score": [0] * batch_size,
     }
-
+    is_finished = torch.full(
+        [batch_size, size],
+        False,
+        dtype=torch.bool,
+        device=device,
+    )
     for step in range(max_output_length):
         # This decides which part of the predicted sentence we feed to the
         # decoder to make the next prediction.
@@ -277,9 +280,9 @@ def beam_search(model: Model, size: int,
             logits, hidden, att_scores, att_vectors = model(
                 return_type="decode",
                 encoder_output=encoder_output,
-                encoder_hidden=None, # used to initialize decoder_hidden only
+                encoder_hidden=None,  # used to initialize decoder_hidden only
                 src_mask=src_mask,
-                trg_input=decoder_input, #trg_embed = embed(decoder_input)
+                trg_input=decoder_input,  # trg_embed = embed(decoder_input)
                 decoder_hidden=hidden,
                 att_vector=att_vectors,
                 unroll_steps=1,
@@ -290,7 +293,7 @@ def beam_search(model: Model, size: int,
         # this point, so we only want to know about the last time step.
         if transformer:
             logits = logits[:, -1]  # keep only the last time step
-            hidden = None           # we don't need to keep it for transformer
+            hidden = None  # we don't need to keep it for transformer
 
         # batch*k x trg_vocab
         log_probs = F.log_softmax(logits, dim=-1).squeeze(1)
@@ -317,13 +320,13 @@ def beam_search(model: Model, size: int,
             topk_log_probs = topk_scores.clone()
 
         # reconstruct beam origin and true word ids from flattened order
-        topk_beam_index = topk_ids.floor_divide(trg_vocab_size)
+        topk_beam_index = topk_ids.div(trg_vocab_size, rounding_mode='floor')
         topk_ids = topk_ids.fmod(trg_vocab_size)
 
         # map beam_index to batch_index in the flat representation
         batch_index = (
-            topk_beam_index
-            + beam_offset[:topk_beam_index.size(0)].unsqueeze(1))
+                topk_beam_index
+                + beam_offset[:topk_beam_index.size(0)].unsqueeze(1))
         select_indices = batch_index.view(-1)
 
         # append latest prediction
@@ -331,11 +334,13 @@ def beam_search(model: Model, size: int,
             [alive_seq.index_select(0, select_indices),
              topk_ids.view(-1, 1)], -1)  # batch_size*k x hyp_len
 
-        is_finished = topk_ids.eq(eos_index)
+        is_finished = (topk_ids.eq(eos_index)
+                       | is_finished
+                       | topk_scores.eq(-np.inf))
         if step + 1 == max_output_length:
             is_finished.fill_(True)
-        # end condition is whether the top beam is finished
-        end_condition = is_finished[:, 0].eq(True)
+        # end condition is whether all beams are finished
+        end_condition = is_finished.all(-1)
 
         # save finished hypotheses
         if is_finished.any():
@@ -378,6 +383,7 @@ def beam_search(model: Model, size: int,
             batch_offset = batch_offset.index_select(0, non_finished)
             alive_seq = predictions.index_select(0, non_finished) \
                 .view(-1, alive_seq.size(-1))
+            is_finished = is_finished.index_select(0, non_finished)
 
         # reorder indices, outputs and masks
         select_indices = batch_index.view(-1)
@@ -407,17 +413,15 @@ def beam_search(model: Model, size: int,
         return filled
 
     # from results to stacked outputs
-    assert n_best == 1
-    # only works for n_best=1 for now
-    final_outputs = pad_and_stack_hyps([r[0].cpu().numpy() for r in
-                                        results["predictions"]],
-                                       pad_value=pad_index)
-
+    final_outputs = pad_and_stack_hyps(
+        [u.cpu().numpy() for r in results["predictions"] for u in r],
+        pad_value=pad_index)
     return final_outputs, None
 
 
 def run_batch(model: Model, batch: Batch, max_output_length: int,
-              beam_size: int, beam_alpha: float) -> (np.array, np.array):
+              beam_size: int, beam_alpha: float,
+              n_best: int = 1) -> (np.array, np.array):
     """
     Get outputs and attentions scores for a given batch
 
@@ -426,6 +430,7 @@ def run_batch(model: Model, batch: Batch, max_output_length: int,
     :param max_output_length: maximum length of hypotheses
     :param beam_size: size of the beam for beam search, if 0 use greedy
     :param beam_alpha: alpha value for beam search
+    :param n_best: candidates to return
     :return: stacked_output: hypotheses for batch,
         stacked_attention_scores: attention scores for batch
     """
@@ -454,6 +459,7 @@ def run_batch(model: Model, batch: Batch, max_output_length: int,
             encoder_hidden=encoder_hidden,
             src_mask=batch.src_mask,
             max_output_length=max_output_length,
-            alpha=beam_alpha)
+            alpha=beam_alpha,
+            n_best=n_best)
 
     return stacked_output, stacked_attention_scores
