@@ -1,31 +1,31 @@
 # coding: utf-8
-
 """
 Training module
 """
 
 import argparse
+import heapq
 import time
 import shutil
-from typing import List
+from typing import List, Tuple
 import logging
 import os
 import sys
-import queue
-
+from pathlib import Path
 import numpy as np
 
 import torch
 from torch import Tensor
 from torch.utils.tensorboard import SummaryWriter
 
-from torchtext.data import Dataset
+# pylint: disable=no-name-in-module
+from torchtext.legacy.data import Dataset
 
 from joeynmt.model import build_model
 from joeynmt.batch import Batch
 from joeynmt.helpers import log_data_info, load_config, log_cfg, \
     store_attention_plots, load_checkpoint, make_model_dir, \
-    make_logger, set_seed, symlink_update, ConfigurationError
+    make_logger, set_seed, symlink_update, delete_ckpt, ConfigurationError
 from joeynmt.model import Model, _DataParallel
 from joeynmt.prediction import validate_on_data
 from joeynmt.loss import XentLoss
@@ -50,6 +50,7 @@ class TrainManager:
     """ Manages training loop, validations, learning rate scheduling
     and early stopping."""
 
+    # pylint: disable=too-many-statements
     def __init__(self, model: Model, config: dict,
                  batch_class: Batch = Batch) -> None:
         """
@@ -67,9 +68,10 @@ class TrainManager:
         assert os.path.exists(self.model_dir)
 
         self.logging_freq = train_config.get("logging_freq", 100)
-        self.valid_report_file = "{}/validations.txt".format(self.model_dir)
-        self.tb_writer = SummaryWriter(
-            log_dir=self.model_dir + "/tensorboard/")
+        self.valid_report_file = f"{self.model_dir}/validations.txt"
+        self.tb_writer = SummaryWriter(log_dir=self.model_dir + "/tensorboard/")
+
+        self.save_latest_checkpoint = train_config.get("save_latest_ckpt", True)
 
         # model
         self.model = model
@@ -92,16 +94,23 @@ class TrainManager:
         self.optimizer = build_optimizer(config=train_config,
                                          parameters=model.parameters())
 
+        # save/delete checkpoints
+        self.num_ckpts = train_config.get("keep_best_ckpts", 5)
+        self.ckpt_queue: List[Tuple[float, Path]] = [] # heap queue
+        # backward compatibility
+        keep_last_ckpts = train_config.get("keep_last_ckpts", None)
+        if keep_last_ckpts is not None:
+            self.num_ckpts = keep_last_ckpts
+            logger.warning("`keep_last_ckpts` option is outdated. "
+                           "Please use `keep_best_ckpts`, instead.")
+
         # validation & early stopping
         self.validation_freq = train_config.get("validation_freq", 1000)
         self.log_valid_sents = train_config.get("print_valid_sents", [0, 1, 2])
-        self.ckpt_queue = queue.Queue(
-            maxsize=train_config.get("keep_last_ckpts", 5))
         self.eval_metric = train_config.get("eval_metric", "bleu")
-        if self.eval_metric not in ['bleu',
-                                    'chrf',
-                                    'token_accuracy',
-                                    'sequence_accuracy']:
+        if self.eval_metric not in [
+                'bleu', 'chrf', 'token_accuracy', 'sequence_accuracy'
+        ]:
             raise ConfigurationError("Invalid setting for 'eval_metric', "
                                      "valid options: 'bleu', 'chrf', "
                                      "'token_accuracy', 'sequence_accuracy'.")
@@ -115,8 +124,9 @@ class TrainManager:
         if self.early_stopping_metric in ["ppl", "loss"]:
             self.minimize_metric = True
         elif self.early_stopping_metric == "eval_metric":
-            if self.eval_metric in ["bleu", "chrf",
-                                    "token_accuracy", "sequence_accuracy"]:
+            if self.eval_metric in [
+                    "bleu", "chrf", "token_accuracy", "sequence_accuracy"
+            ]:
                 self.minimize_metric = False
             # eval metric that has to get minimized (not yet implemented)
             else:
@@ -150,7 +160,11 @@ class TrainManager:
                                      "Valid options: 'word', 'bpe', 'char'.")
         self.shuffle = train_config.get("shuffle", True)
         self.epochs = train_config["epochs"]
+        self.max_updates = train_config.get("updates", np.inf)
         self.batch_size = train_config["batch_size"]
+        # Placeholder so that we can use the train_iter in other functions.
+        self.train_iter = None
+        self.train_iter_state = None
         # per-device batch_size = self.batch_size // self.n_gpu
         self.batch_type = train_config.get("batch_type", "sentence")
         self.eval_batch_size = train_config.get("eval_batch_size",
@@ -175,37 +189,39 @@ class TrainManager:
         self.fp16 = train_config.get("fp16", False)
         if self.fp16:
             if 'apex' not in sys.modules:
-                raise ImportError(
-                    "Please install apex from "
-                    "https://www.github.com/nvidia/apex "
-                    "to use fp16 training.") from no_apex
-            self.model, self.optimizer = amp.initialize(
-                self.model, self.optimizer, opt_level='O1')
+                raise ImportError("Please install apex from "
+                                  "https://www.github.com/nvidia/apex "
+                                  "to use fp16 training.") from no_apex  # pylint: disable=used-before-assignment
+            self.model, self.optimizer = amp.initialize(self.model,
+                                                        self.optimizer,
+                                                        opt_level='O1')
             # opt level: one of {"O0", "O1", "O2", "O3"}
             # see https://nvidia.github.io/apex/amp.html#opt-levels
 
         # initialize training statistics
         self.stats = self.TrainStatistics(
             steps=0,
-            stop=False,
+            is_min_lr=False,
+            is_max_update=False,
             total_tokens=0,
             best_ckpt_iter=0,
             best_ckpt_score=np.inf if self.minimize_metric else -np.inf,
-            minimize_metric=self.minimize_metric
-        )
+            minimize_metric=self.minimize_metric)
 
         # model parameters
         if "load_model" in train_config.keys():
-            self.init_from_checkpoint(train_config["load_model"],
+            self.init_from_checkpoint(
+                train_config["load_model"],
                 reset_best_ckpt=train_config.get("reset_best_ckpt", False),
                 reset_scheduler=train_config.get("reset_scheduler", False),
-                reset_optimizer=train_config.get("reset_optimizer", False))
+                reset_optimizer=train_config.get("reset_optimizer", False),
+                reset_iter_state=train_config.get("reset_iter_state", False))
 
         # multi-gpu training (should be after apex fp16 initialization)
         if self.n_gpu > 1:
             self.model = _DataParallel(self.model)
 
-    def _save_checkpoint(self) -> None:
+    def _save_checkpoint(self, new_best: bool, score: float) -> None:
         """
         Save the model's current parameters and the training state to a
         checkpoint.
@@ -215,45 +231,81 @@ class TrainManager:
         the best checkpoint score and iteration so far,
         and optimizer and scheduler states.
 
+        :param new_best: This boolean signals which symlink we will use for the
+                         new checkpoint. If it is true, we update best.ckpt.
+        :param score: Validation score which is used as key of heap queue.
         """
-        model_path = "{}/{}.ckpt".format(self.model_dir, self.stats.steps)
+        model_path = Path(self.model_dir) / f"{self.stats.steps}.ckpt"
         model_state_dict = self.model.module.state_dict() \
             if isinstance(self.model, torch.nn.DataParallel) \
             else self.model.state_dict()
         state = {
-            "steps": self.stats.steps,
-            "total_tokens": self.stats.total_tokens,
-            "best_ckpt_score": self.stats.best_ckpt_score,
-            "best_ckpt_iteration": self.stats.best_ckpt_iter,
-            "model_state": model_state_dict,
-            "optimizer_state": self.optimizer.state_dict(),
-            "scheduler_state": self.scheduler.state_dict() if
-            self.scheduler is not None else None,
-            'amp_state': amp.state_dict() if self.fp16 else None
+            "steps":
+            self.stats.steps,
+            "total_tokens":
+            self.stats.total_tokens,
+            "best_ckpt_score":
+            self.stats.best_ckpt_score,
+            "best_ckpt_iteration":
+            self.stats.best_ckpt_iter,
+            "model_state":
+            model_state_dict,
+            "optimizer_state":
+            self.optimizer.state_dict(),
+            "scheduler_state":
+            self.scheduler.state_dict() if self.scheduler is not None else None,
+            'amp_state':
+            amp.state_dict() if self.fp16 else None,
+            "train_iter_state":
+            self.train_iter.state_dict()
         }
-        torch.save(state, model_path)
-        if self.ckpt_queue.full():
-            to_delete = self.ckpt_queue.get()  # delete oldest ckpt
-            try:
-                os.remove(to_delete)
-            except FileNotFoundError:
-                logger.warning("Wanted to delete old checkpoint %s but "
-                               "file does not exist.", to_delete)
+        torch.save(state, model_path.as_posix())
 
-        self.ckpt_queue.put(model_path)
+        # update symlink
+        symlink_target = Path(f"{self.stats.steps}.ckpt")
+        # last symlink
+        last_path = Path(self.model_dir) / "latest.ckpt"
+        prev_path = symlink_update(symlink_target, last_path) # update always
+        # best symlink
+        best_path = Path(self.model_dir) / "best.ckpt"
+        if new_best:
+            prev_path = symlink_update(symlink_target, best_path)
+            assert best_path.resolve().stem == str(self.stats.best_ckpt_iter)
 
-        best_path = "{}/best.ckpt".format(self.model_dir)
-        try:
-            # create/modify symbolic link for best checkpoint
-            symlink_update("{}.ckpt".format(self.stats.steps), best_path)
-        except OSError:
-            # overwrite best.ckpt
-            torch.save(state, best_path)
+        # push to and pop from the heap queue
+        if self.num_ckpts > 0:
+            to_delete = None
+            if len(self.ckpt_queue) < self.num_ckpts:   # no pop, push only
+                heapq.heappush(self.ckpt_queue, (score, model_path))
+            else:   # push + pop the worst one in the queue
+                if self.minimize_metric:
+                    # pylint: disable=protected-access
+                    heapq._heapify_max(self.ckpt_queue)
+                    to_delete = heapq._heappop_max(self.ckpt_queue)
+                    heapq.heappush(self.ckpt_queue, (score, model_path))
+                    # pylint: enable=protected-access
+                else:
+                    to_delete = heapq.heappushpop(self.ckpt_queue,
+                                                  (score, model_path))
+                assert to_delete[1] != model_path   # don't delete the last ckpt
 
-    def init_from_checkpoint(self, path: str,
+            if to_delete is not None \
+                    and to_delete[1].stem != best_path.resolve().stem:
+                delete_ckpt(to_delete[1])           # don't delete the best ckpt
+
+            assert len(self.ckpt_queue) <= self.num_ckpts
+
+            # remove old symlink target if not in queue after push/pop
+            if prev_path is not None and \
+                    prev_path.stem not in [c[1].stem for c in self.ckpt_queue]:
+                delete_ckpt(prev_path)
+
+    def init_from_checkpoint(self,
+                             path: str,
                              reset_best_ckpt: bool = False,
                              reset_scheduler: bool = False,
-                             reset_optimizer: bool = False) -> None:
+                             reset_optimizer: bool = False,
+                             reset_iter_state: bool = False) -> None:
         """
         Initialize the trainer from a given checkpoint file.
 
@@ -268,6 +320,8 @@ class TrainManager:
                                 use the one stored in the checkpoint.
         :param reset_optimizer: reset the optimizer, and do not use the one
                                 stored in the checkpoint.
+        :param reset_iter_state: reset the sampler's internal state and do not
+                                use the one stored in the checkpoint.
         """
         logger.info("Loading model from %s", path)
         model_checkpoint = load_checkpoint(path=path, use_cuda=self.use_cuda)
@@ -298,6 +352,12 @@ class TrainManager:
         else:
             logger.info("Reset tracking of the best checkpoint.")
 
+        if not reset_iter_state:
+            assert 'train_iter_state' in model_checkpoint
+            self.train_iter_state = model_checkpoint["train_iter_state"]
+        else:
+            logger.info("Reset train data iterator.")
+
         # move parameters to cuda
         if self.use_cuda:
             self.model.to(self.device)
@@ -317,10 +377,14 @@ class TrainManager:
         :param train_data: training data
         :param valid_data: validation data
         """
-        train_iter = make_data_iter(train_data,
-                                    batch_size=self.batch_size,
-                                    batch_type=self.batch_type,
-                                    train=True, shuffle=self.shuffle)
+        self.train_iter = make_data_iter(train_data,
+                                         batch_size=self.batch_size,
+                                         batch_type=self.batch_type,
+                                         train=True,
+                                         shuffle=self.shuffle)
+
+        if self.train_iter_state is not None:
+            self.train_iter.load_state_dict(self.train_iter_state)
 
         #################################################################
         # simplify accumulation logic:
@@ -329,7 +393,7 @@ class TrainManager:
         #     self.model.zero_grad()
         #     epoch_loss = 0.0
         #     batch_loss = 0.0
-        #     for i, batch in enumerate(iter(train_iter)):
+        #     for i, batch in enumerate(iter(self.train_iter)):
         #
         #         # gradient accumulation:
         #         # loss.backward() inside _train_step()
@@ -353,9 +417,9 @@ class TrainManager:
             "\t16-bits training: %r\n"
             "\tgradient accumulation: %d\n"
             "\tbatch size per device: %d\n"
-            "\ttotal batch size (w. parallel & accumulation): %d",
-            self.device, self.n_gpu, self.fp16, self.batch_multiplier,
-            self.batch_size//self.n_gpu if self.n_gpu > 1 else self.batch_size,
+            "\ttotal batch size (w. parallel & accumulation): %d", self.device,
+            self.n_gpu, self.fp16, self.batch_multiplier, self.batch_size //
+            self.n_gpu if self.n_gpu > 1 else self.batch_size,
             self.batch_size * self.batch_multiplier)
 
         for epoch_no in range(self.epochs):
@@ -374,7 +438,7 @@ class TrainManager:
             epoch_loss = 0
             batch_loss = 0
 
-            for i, batch in enumerate(iter(train_iter)):
+            for i, batch in enumerate(iter(self.train_iter)):
                 # create a Batch object from torchtext batch
                 batch = self.batch_class(batch, self.model.pad_index,
                                          use_cuda=self.use_cuda)
@@ -405,6 +469,8 @@ class TrainManager:
 
                     # increment step counter
                     self.stats.steps += 1
+                    if self.stats.steps >= self.max_updates:
+                        self.stats.is_max_update = True
 
                     # log learning progress
                     if self.stats.steps % self.logging_freq == 0:
@@ -414,8 +480,8 @@ class TrainManager:
                         elapsed_tokens = self.stats.total_tokens - start_tokens
                         logger.info(
                             "Epoch %3d, Step: %8d, Batch Loss: %12.6f, "
-                            "Tokens per Sec: %8.0f, Lr: %.6f",
-                            epoch_no + 1, self.stats.steps, batch_loss,
+                            "Tokens per Sec: %8.0f, Lr: %.6f", epoch_no + 1,
+                            self.stats.steps, batch_loss,
                             elapsed_tokens / elapsed,
                             self.optimizer.param_groups[0]["lr"])
                         start = time.time()
@@ -423,24 +489,34 @@ class TrainManager:
                         start_tokens = self.stats.total_tokens
 
                     # Only add complete loss of full mini-batch to epoch_loss
-                    epoch_loss += batch_loss    # accumulate epoch_loss
-                    batch_loss = 0              # rest batch_loss
+                    epoch_loss += batch_loss  # accumulate epoch_loss
+                    batch_loss = 0  # rest batch_loss
 
                     # validate on the entire dev set
                     if self.stats.steps % self.validation_freq == 0:
                         valid_duration = self._validate(valid_data, epoch_no)
                         total_valid_duration += valid_duration
 
-                if self.stats.stop:
+                    # check current_lr
+                    current_lr = self.optimizer.param_groups[0]['lr']
+                    if current_lr < self.learning_rate_min:
+                        self.stats.is_min_lr = True
+
+                    self.tb_writer.add_scalar("train/learning_rate",
+                                              current_lr, self.stats.steps)
+
+                if self.stats.is_min_lr or self.stats.is_max_update:
                     break
-            if self.stats.stop:
-                logger.info(
-                    'Training ended since minimum lr %f was reached.',
-                    self.learning_rate_min)
+
+            if self.stats.is_min_lr or self.stats.is_max_update:
+                log_str = f"minimum lr {self.learning_rate_min}" \
+                    if self.stats.is_min_lr \
+                    else f"maximum num. of updates {self.max_updates}"
+                logger.info("Training ended since %s was reached.", log_str)
                 break
 
-            logger.info('Epoch %3d: total training loss %.2f',
-                        epoch_no + 1, epoch_loss)
+            logger.info('Epoch %3d: total training loss %.2f', epoch_no + 1,
+                        epoch_loss)
         else:
             logger.info('Training ended after %3d epochs.', epoch_no + 1)
         logger.info('Best validation result (greedy) at step %8d: %6.2f %s.',
@@ -474,9 +550,8 @@ class TrainManager:
         elif self.normalization == "none":
             normalizer = 1
         else:
-            raise NotImplementedError(
-                "Only normalize by 'batch' or 'tokens' "
-                "or summation of loss 'none' implemented")
+            raise NotImplementedError("Only normalize by 'batch' or 'tokens' "
+                                      "or summation of loss 'none' implemented")
 
         norm_batch_loss = batch_loss / normalizer
 
@@ -521,12 +596,12 @@ class TrainManager:
                 n_gpu=self.n_gpu
             )
 
-        self.tb_writer.add_scalar(
-            "valid/valid_loss", valid_loss, self.stats.steps)
-        self.tb_writer.add_scalar(
-            "valid/valid_score", valid_score, self.stats.steps)
-        self.tb_writer.add_scalar(
-            "valid/valid_ppl", valid_ppl, self.stats.steps)
+        self.tb_writer.add_scalar("valid/valid_loss", valid_loss,
+                                  self.stats.steps)
+        self.tb_writer.add_scalar("valid/valid_score", valid_score,
+                                  self.stats.steps)
+        self.tb_writer.add_scalar("valid/valid_ppl", valid_ppl,
+                                  self.stats.steps)
 
         if self.early_stopping_metric == "loss":
             ckpt_score = valid_loss
@@ -535,61 +610,65 @@ class TrainManager:
         else:
             ckpt_score = valid_score
 
-        new_best = False
-        if self.stats.is_best(ckpt_score):
-            self.stats.best_ckpt_score = ckpt_score
-            self.stats.best_ckpt_iter = self.stats.steps
-            logger.info('Hooray! New best validation result [%s]!',
-                        self.early_stopping_metric)
-            if self.ckpt_queue.maxsize > 0:
-                logger.info("Saving new checkpoint.")
-                new_best = True
-                self._save_checkpoint()
-
         if self.scheduler is not None \
                 and self.scheduler_step_at == "validation":
             self.scheduler.step(ckpt_score)
 
-        # append to validation report
-        self._add_report(
-            valid_score=valid_score, valid_loss=valid_loss,
-            valid_ppl=valid_ppl, eval_metric=self.eval_metric,
-            new_best=new_best)
+        # update new best
+        new_best = self.stats.is_best(ckpt_score)
+        if new_best:
+            self.stats.best_ckpt_score = ckpt_score
+            self.stats.best_ckpt_iter = self.stats.steps
+            logger.info('Hooray! New best validation result [%s]!',
+                        self.early_stopping_metric)
 
-        self._log_examples(
-            sources_raw=[v for v in valid_sources_raw],
-            sources=valid_sources,
-            hypotheses_raw=valid_hypotheses_raw,
-            hypotheses=valid_hypotheses,
-            references=valid_references
-        )
+        # save checkpoints
+        is_better = self.stats.is_better(ckpt_score, self.ckpt_queue) \
+            if len(self.ckpt_queue) > 0 else True
+        if self.num_ckpts < 0 or is_better:
+            self._save_checkpoint(new_best, ckpt_score)
+
+        # append to validation report
+        self._add_report(valid_score=valid_score,
+                         valid_loss=valid_loss,
+                         valid_ppl=valid_ppl,
+                         eval_metric=self.eval_metric,
+                         new_best=new_best)
+
+        self._log_examples(sources_raw=[v for v in valid_sources_raw],
+                           sources=valid_sources,
+                           hypotheses_raw=valid_hypotheses_raw,
+                           hypotheses=valid_hypotheses,
+                           references=valid_references)
 
         valid_duration = time.time() - valid_start_time
         logger.info(
             'Validation result (greedy) at epoch %3d, '
             'step %8d: %s: %6.2f, loss: %8.4f, ppl: %8.4f, '
-            'duration: %.4fs', epoch_no + 1, self.stats.steps,
-            self.eval_metric, valid_score, valid_loss,
-            valid_ppl, valid_duration)
+            'duration: %.4fs', epoch_no + 1, self.stats.steps, self.eval_metric,
+            valid_score, valid_loss, valid_ppl, valid_duration)
 
         # store validation set outputs
         self._store_outputs(valid_hypotheses)
 
         # store attention plots for selected valid sentences
         if valid_attention_scores:
-            store_attention_plots(
-                attentions=valid_attention_scores,
-                targets=valid_hypotheses_raw,
-                sources=[s for s in valid_data.src],
-                indices=self.log_valid_sents,
-                output_prefix="{}/att.{}".format(
-                    self.model_dir, self.stats.steps),
-                tb_writer=self.tb_writer, steps=self.stats.steps)
+            store_attention_plots(attentions=valid_attention_scores,
+                                  targets=valid_hypotheses_raw,
+                                  sources=[s for s in valid_data.src],
+                                  indices=self.log_valid_sents,
+                                  output_prefix=f"{self.model_dir}/" \
+                                          f"att.{self.stats.steps}",
+                                  tb_writer=self.tb_writer,
+                                  steps=self.stats.steps)
 
         return valid_duration
 
-    def _add_report(self, valid_score: float, valid_ppl: float,
-                    valid_loss: float, eval_metric: str,
+    def _add_report(self,
+                    valid_score: float,
+                    valid_ppl: float,
+                    valid_loss: float,
+                    eval_metric: str,
                     new_best: bool = False) -> None:
         """
         Append a one-line report to validation logging file.
@@ -600,20 +679,15 @@ class TrainManager:
         :param eval_metric: evaluation metric, e.g. "bleu"
         :param new_best: whether this is a new best model
         """
-        current_lr = -1
         # ignores other param groups for now
-        for param_group in self.optimizer.param_groups:
-            current_lr = param_group['lr']
+        current_lr = self.optimizer.param_groups[0]['lr']
 
-        if current_lr < self.learning_rate_min:
-            self.stats.stop = True
-
-        with open(self.valid_report_file, 'a') as opened_file:
+        with open(self.valid_report_file, 'a', encoding="utf-8") as opened_file:
             opened_file.write(
-                "Steps: {}\tLoss: {:.5f}\tPPL: {:.5f}\t{}: {:.5f}\t"
-                "LR: {:.8f}\t{}\n".format(
-                    self.stats.steps, valid_loss, valid_ppl, eval_metric,
-                    valid_score, current_lr, "*" if new_best else ""))
+                f"Steps: {self.stats.steps}\tLoss: {valid_loss:.5f}\t" \
+                        f"PPL: {valid_ppl:.5f}\t{eval_metric}: "\
+                        f"{valid_score:.5f}\tLR: {current_lr:.8f}\t" \
+                        f"{'*' if new_best else ''}\n")
 
     def _log_parameters_list(self) -> None:
         """
@@ -623,12 +697,15 @@ class TrainManager:
                                   self.model.parameters())
         n_params = sum([np.prod(p.size()) for p in model_parameters])
         logger.info("Total params: %d", n_params)
-        trainable_params = [n for (n, p) in self.model.named_parameters()
-                            if p.requires_grad]
+        trainable_params = [
+            n for (n, p) in self.model.named_parameters() if p.requires_grad
+        ]
         logger.debug("Trainable parameters: %s", sorted(trainable_params))
         assert trainable_params
 
-    def _log_examples(self, sources: List[str], hypotheses: List[str],
+    def _log_examples(self,
+                      sources: List[str],
+                      hypotheses: List[str],
                       references: List[str],
                       sources_raw: List[List[str]] = None,
                       hypotheses_raw: List[List[str]] = None,
@@ -667,22 +744,30 @@ class TrainManager:
 
         :param hypotheses: list of strings
         """
-        current_valid_output_file = "{}/{}.hyps".format(self.model_dir,
-                                                        self.stats.steps)
-        with open(current_valid_output_file, 'w') as opened_file:
+        current_valid_output_file = f"{self.model_dir}/" \
+                f"{self.stats.steps}.hyps"
+        with open(current_valid_output_file, 'w', encoding="utf-8") \
+                as opened_file:
             for hyp in hypotheses:
-                opened_file.write("{}\n".format(hyp))
+                opened_file.write(f"{hyp}\n")
 
     class TrainStatistics:
-        def __init__(self, steps: int = 0, stop: bool = False,
-                     total_tokens: int = 0, best_ckpt_iter: int = 0,
+        def __init__(self,
+                     steps: int = 0,
+                     is_min_lr: bool = False,
+                     is_max_update: bool = False,
+                     total_tokens: int = 0,
+                     best_ckpt_iter: int = 0,
                      best_ckpt_score: float = np.inf,
                      minimize_metric: bool = True) -> None:
             # global update step counter
             self.steps = steps
             # stop training if this flag is True
             # by reaching learning rate minimum
-            self.stop = stop
+            self.is_min_lr = is_min_lr
+            # stop training if this flag is True
+            # by reaching max num of updates
+            self.is_max_update = is_max_update
             # number of total tokens seen so far
             self.total_tokens = total_tokens
             # store iteration point of best ckpt
@@ -699,19 +784,29 @@ class TrainManager:
                 is_best = score > self.best_ckpt_score
             return is_best
 
+        def is_better(self, score: float, heap_queue: list):
+            assert len(heap_queue) > 0
+            if self.minimize_metric:
+                is_better = score < heapq.nlargest(1, heap_queue)[0][0]
+            else:
+                is_better = score > heapq.nsmallest(1, heap_queue)[0][0]
+            return is_better
 
-def train(cfg_file: str) -> None:
+
+def train(cfg_file: str, skip_test: bool = False) -> None:
     """
     Main training function. After training, also test on test data if given.
 
     :param cfg_file: path to configuration yaml file
+    :param skip_test: whether a test should be run or not after training
     """
     cfg = load_config(cfg_file)
 
     # make logger
     model_dir = make_model_dir(cfg["training"]["model_dir"],
-                   overwrite=cfg["training"].get("overwrite", False))
-    _ = make_logger(model_dir, mode="train")    # version string returned
+                               overwrite=cfg["training"].get(
+                                   "overwrite", False))
+    _ = make_logger(model_dir, mode="train")  # version string returned
     # TODO: save version number in model checkpoints
 
     # set the random seed
@@ -733,34 +828,47 @@ def train(cfg_file: str) -> None:
     # log all entries of config
     log_cfg(cfg)
 
-    log_data_info(train_data=train_data, valid_data=dev_data,
-                  test_data=test_data, src_vocab=src_vocab, trg_vocab=trg_vocab)
+    log_data_info(train_data=train_data,
+                  valid_data=dev_data,
+                  test_data=test_data,
+                  src_vocab=src_vocab,
+                  trg_vocab=trg_vocab)
 
     logger.info(str(model))
 
     # store the vocabs
-    src_vocab_file = "{}/src_vocab.txt".format(cfg["training"]["model_dir"])
+    src_vocab_file = f"{model_dir}/src_vocab.txt"
     src_vocab.to_file(src_vocab_file)
-    trg_vocab_file = "{}/trg_vocab.txt".format(cfg["training"]["model_dir"])
+    trg_vocab_file = f"{model_dir}/trg_vocab.txt"
     trg_vocab.to_file(trg_vocab_file)
 
     # train the model
     trainer.train_and_validate(train_data=train_data, valid_data=dev_data)
 
-    # predict with the best model on validation and test
-    # (if test data is available)
-    ckpt = "{}/{}.ckpt".format(model_dir, trainer.stats.best_ckpt_iter)
-    output_name = "{:08d}.hyps".format(trainer.stats.best_ckpt_iter)
-    output_path = os.path.join(model_dir, output_name)
-    datasets_to_test = {"dev": dev_data, "test": test_data,
-                        "src_vocab": src_vocab, "trg_vocab": trg_vocab}
-    test(cfg_file, ckpt=ckpt, output_path=output_path,
-         datasets=datasets_to_test)
-
+    if not skip_test:
+        # predict with the best model on validation and test
+        # (if test data is available)
+        ckpt = f"{model_dir}/{trainer.stats.best_ckpt_iter}.ckpt"
+        output_name = f"{trainer.stats.best_ckpt_iter:08d}.hyps"
+        output_path = os.path.join(model_dir, output_name)
+        datasets_to_test = {
+            "dev": dev_data,
+            "test": test_data,
+            "src_vocab": src_vocab,
+            "trg_vocab": trg_vocab
+        }
+        test(cfg_file,
+             ckpt=ckpt,
+             output_path=output_path,
+             datasets=datasets_to_test)
+    else:
+        logger.info("Skipping test after training")
 
 if __name__ == "__main__":
     parser = argparse.ArgumentParser('Joey-NMT')
-    parser.add_argument("config", default="configs/default.yaml", type=str,
+    parser.add_argument("config",
+                        default="configs/default.yaml",
+                        type=str,
                         help="Training configuration file (yaml).")
     args = parser.parse_args()
     train(cfg_file=args.config)
