@@ -7,7 +7,8 @@ from collections import defaultdict, Counter
 from typing import List
 import numpy as np
 
-from torchtext.data import Dataset
+# pylint: disable=no-name-in-module
+from torchtext.legacy.data import Dataset
 
 from joeynmt.constants import UNK_TOKEN, DEFAULT_UNK_ID, \
     EOS_TOKEN, BOS_TOKEN, PAD_TOKEN
@@ -58,7 +59,7 @@ class Vocabulary:
         :param file: path to file where the vocabulary is loaded from
         """
         tokens = []
-        with open(file, "r") as open_file:
+        with open(file, "r", encoding='utf-8') as open_file:
             for line in open_file:
                 tokens.append(line.strip("\n"))
         self._from_list(tokens)
@@ -72,9 +73,9 @@ class Vocabulary:
 
         :param file: path to file where the vocabulary is written
         """
-        with open(file, "w") as open_file:
+        with open(file, "w", encoding="utf-8") as open_file:
             for t in self.itos:
-                open_file.write("{}\n".format(t))
+                open_file.write(f"{t}\n")
 
     def add_tokens(self, tokens: List[str]) -> None:
         """
